@@ -1,6 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2005, 2006, 2007, 2008, 2009, 2010, 2011 Free Software
-   Foundation, Inc.
+   Copyright (C) 2005-2012, 2014-2016 Free Software Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -18,44 +17,10 @@
 /* MH burst command */
 
 #include <mh.h>
-#define obstack_chunk_alloc malloc
-#define obstack_chunk_free free
-#include <obstack.h>
 
-static char doc[] = N_("GNU MH burst")"\v"
-N_("Options marked with `*' are not yet implemented.\n\
-Use -help to obtain the list of traditional MH options.");
+static char prog_doc[] = N_("Explode digests into messages");
 static char args_doc[] = N_("[MSGLIST]");
 
-/* GNU options */
-static struct argp_option options[] = {
-  {"folder",        ARG_FOLDER,        N_("FOLDER"), 0,
-   N_("specify folder to operate upon")},
-  {"inplace",      ARG_INPLACE,      N_("BOOL"), OPTION_ARG_OPTIONAL,
-   N_("replace the source message with the table of contents, insert extracted messages after it") },
-  {"noinplace",    ARG_NOINPLACE,    0, OPTION_HIDDEN, ""},
-  {"quiet",        ARG_QUIET,        N_("BOOL"), OPTION_ARG_OPTIONAL,
-   N_("be quiet about the messages that are not in digest format") },
-  {"noquiet",      ARG_NOQUIET,      0, OPTION_HIDDEN, ""},
-  {"verbose",      ARG_VERBOSE,      N_("BOOL"), OPTION_ARG_OPTIONAL,
-   N_("verbosely list the actions taken") },
-  {"noverbose",    ARG_NOVERBOSE,    0, OPTION_HIDDEN, ""},
-  {"recursive",    ARG_RECURSIVE,    N_("BOOL"), OPTION_ARG_OPTIONAL,
-   N_("recursively expand MIME messages") },
-  {"norecursive",  ARG_NORECURSIVE,  0, OPTION_HIDDEN, ""},
-  {"length",       ARG_LENGTH,       N_("NUMBER"), 0,
-   N_("set minimal length of digest encapsulation boundary (default 1)") },
-  { NULL }
-};
-
-/* Traditional MH options */
-struct mh_option mh_option[] = {
-  { "inplace",    MH_OPT_BOOL },
-  { "quiet",      MH_OPT_BOOL },
-  { "verbose",    MH_OPT_BOOL },
-  { NULL }
-};
-
 /* Command line switches */
 int inplace; 
 int quiet;
@@ -65,59 +30,24 @@ int eb_min_length = 1;  /* Minimal length of encapsulation boundary */
 
 #define VERBOSE(c) do { if (verbose) { printf c; putchar ('\n'); } } while (0)
 
-static error_t
-opt_handler (int key, char *arg, struct argp_state *state)
-{
-  switch (key)
-    {
-    case ARG_FOLDER: 
-      mh_set_current_folder (arg);
-      break;
-
-    case ARG_INPLACE:
-      inplace = is_true (arg);
-      break;
-
-    case ARG_NOINPLACE:
-      inplace = 0;
-      break;
-
-    case ARG_LENGTH:
-      eb_min_length = strtoul (arg, NULL, 0);
-      if (eb_min_length == 0)
-	eb_min_length = 1;
-      break;
-      
-    case ARG_VERBOSE:
-      verbose = is_true (arg);
-      break;
-
-    case ARG_NOVERBOSE:
-      verbose = 0;
-      break;
-
-    case ARG_RECURSIVE:
-      recursive = is_true (arg);
-      break;
-      
-    case ARG_NORECURSIVE:
-      recursive = 0;
-      break;
-      
-    case ARG_QUIET:
-      quiet = is_true (arg);
-      break;
-
-    case ARG_NOQUIET:
-      quiet = 0;
-      break;
-      
-    default:
-      return ARGP_ERR_UNKNOWN;
-    }
-  return 0;
-}
-
+static struct mu_option options[] = {
+  { "inplace",     0,      NULL, MU_OPTION_DEFAULT,
+   N_("replace the source message with the table of contents, insert extracted messages after it"),
+    mu_c_bool, &inplace },
+  { "quiet",       0,      NULL, MU_OPTION_DEFAULT,
+    N_("be quiet about the messages that are not in digest format"),
+    mu_c_bool, &quiet },
+  { "verbose",     0,      NULL, MU_OPTION_DEFAULT,
+    N_("verbosely list the actions taken"),
+    mu_c_bool, &verbose },
+  { "recursive",   0,      NULL, MU_OPTION_DEFAULT,
+    N_("recursively expand MIME messages"),
+    mu_c_bool, &recursive },
+  { "length",      0, N_("NUM"), MU_OPTION_DEFAULT,
+    N_("set minimal length of digest encapsulation boundary (default 1)"),
+    mu_c_int, &eb_min_length },
+  MU_OPTION_END
+};
 
 /* General-purpose data structures */
 struct burst_map
@@ -135,7 +65,7 @@ struct burst_map map;        /* Currently built map */
 struct burst_map *burst_map; /* Finished burst map */
 size_t burst_count;          /* Number of items in burst_map */
 mu_mailbox_t tmpbox;         /* Temporary mailbox */
-struct obstack stk;          /* Stack for building burst_map, etc. */
+mu_opool_t pool;             /* Object pool for building burst_map, etc. */
 
 static int burst_or_copy (mu_message_t msg, int recursive, int copy);
 
@@ -578,8 +508,8 @@ burst_or_copy (mu_message_t msg, int recursive, int copy)
   return 1;
 }
 
-void
-burst (mu_mailbox_t mbox, mu_message_t msg, size_t num, void *data)
+int
+burst (size_t num, mu_message_t msg, void *data)
 {
   memset (&map, 0, sizeof (map));
   mh_message_number (msg, &map.msgno);
@@ -593,42 +523,44 @@ burst (mu_mailbox_t mbox, mu_message_t msg, size_t num, void *data)
 	       mu_umaxtostr (1, num)));
       if (inplace)
 	{
-	  obstack_grow (&stk, &map, sizeof map);
+	  mu_opool_append (pool, &map, sizeof map);
 	  burst_count++;
 	}
     }
   else if (!quiet)
     mu_error (_("message %s not in digest format"), mu_umaxtostr (0, num));
+  return 0;
 }
 
 
 /* Inplace handling */
-
-void
-burst_rename (mh_msgset_t *ms, size_t lastuid)
+struct rename_env
 {
-  size_t i, j;
+  size_t lastuid;
+  size_t idx;
+};
 
-  VERBOSE ((_("Renaming messages")));
-  j = burst_count - 1;
-  for (i = ms->count; i > 0; i--)
+  
+static int
+_rename (size_t msgno, void *data)
+{
+  struct rename_env *rp = data;
+  
+  if (msgno == burst_map[rp->idx].msgno)
+    {
+      rp->lastuid -= burst_map[rp->idx].count;
+      burst_map[rp->idx].msgno = rp->lastuid;
+      rp->idx--;
+    }
+
+  if (msgno != rp->lastuid)
     {
       const char *from;
       const char *to;
 
-      if (ms->list[i-1] == burst_map[j].msgno)
-	{
-	  lastuid -= burst_map[j].count;
-	  burst_map[j].msgno = lastuid;
-	  j--;
-	}
-
-      if (ms->list[i-1] == lastuid)
-	continue;
-      
-      from = mu_umaxtostr (0, ms->list[i-1]);
-      to   = mu_umaxtostr (1, lastuid);
-      --lastuid;
+      from = mu_umaxtostr (0, msgno);
+      to   = mu_umaxtostr (1, rp->lastuid);
+      --rp->lastuid;
 
       VERBOSE((_("message %s becomes message %s"), from, to));
 	       
@@ -639,6 +571,18 @@ burst_rename (mh_msgset_t *ms, size_t lastuid)
 	  exit (1);
 	}
     }
+  return 0;
+}
+
+void
+burst_rename (mu_msgset_t ms, size_t lastuid)
+{
+  struct rename_env renv;
+
+  VERBOSE ((_("Renaming messages")));
+  renv.lastuid = lastuid;
+  renv.idx = burst_count - 1;
+  mu_msgset_foreach_dir_msguid (ms, 1, _rename, &renv);
 }  
 
 void
@@ -703,24 +647,19 @@ finalize_inplace (size_t lastuid)
 int
 main (int argc, char **argv)
 {
-  int index, rc;
+  int rc;
   mu_mailbox_t mbox;
-  mh_msgset_t msgset;
+  mu_msgset_t msgset;
   const char *tempfolder = mh_global_profile_get ("Temp-Folder", ".temp");
   
-  /* Native Language Support */
-  MU_APP_INIT_NLS ();
-
-  mh_argp_init ();
-  mh_argp_parse (&argc, &argv, 0, options, mh_option, args_doc, doc,
-		 opt_handler, NULL, &index);
-
-  argc -= index;
-  argv += index;
-
+  mh_getopt (&argc, &argv, options, MH_GETOPT_DEFAULT_FOLDER,
+	     args_doc, prog_doc, NULL);
+  if (eb_min_length == 0)
+    eb_min_length = 1;
+  
   VERBOSE ((_("Opening folder `%s'"), mh_current_folder ()));
   mbox = mh_open_folder (mh_current_folder (), MU_STREAM_RDWR);
-  mh_msgset_parse (mbox, &msgset, argc, argv, "cur");
+  mh_msgset_parse (&msgset, mbox, argc, argv, "cur");
 
   if (inplace)
     {
@@ -739,12 +678,12 @@ main (int argc, char **argv)
 	  mu_attribute_set_deleted (attr);
 	}
       mu_mailbox_expunge (tmpbox);
-      obstack_init (&stk);
+      mu_opool_create (&pool, 1);
     }
   else
     tmpbox = mbox;
 
-  rc = mh_iterate (mbox, &msgset, burst, NULL);
+  rc = mu_msgset_foreach_message (msgset, burst, NULL);
   if (rc)
     return rc;
 
@@ -752,22 +691,25 @@ main (int argc, char **argv)
     {
       mu_url_t dst_url = NULL;
       size_t i, next_uid, last_uid;
-      mh_msgset_t ms;
-      char *xargv[2];
+      mu_msgset_t ms;
+      size_t count;
       const char *dir;
       
-      burst_map = obstack_finish (&stk);
+      burst_map = mu_opool_finish (pool, NULL);
 
       mu_mailbox_uidnext (mbox, &next_uid);
       for (i = 0, last_uid = next_uid-1; i < burst_count; i++)
 	last_uid += burst_map[i].count;
       VERBOSE ((_("Estimated last UID: %s"), mu_umaxtostr (0, last_uid)));
 
-      mu_asprintf (&xargv[0], "%s-last", mu_umaxtostr (0, burst_map[0].msgno));
-      xargv[1] = NULL;
-      mh_msgset_parse (mbox, &ms, 1, xargv, NULL);
-      free (xargv[0]);
-      mh_msgset_uids (mbox, &ms);
+      rc = mu_msgset_create (&ms, mbox, MU_MSGSET_NUM);
+      if (rc)
+	{
+	  mu_diag_funcall (MU_DIAG_ERROR, "mu_msgset_create", NULL, rc);
+	  exit (1);
+	}
+      mu_mailbox_messages_count (mbox, &count);
+      mu_msgset_add_range (ms, burst_map[0].msgno, count, MU_MSGSET_NUM);
 	
       mu_mailbox_get_url (mbox, &dst_url);
       mu_url_sget_path (dst_url, &dir);
@@ -779,8 +721,8 @@ main (int argc, char **argv)
 	}
       mu_mailbox_close (mbox);
 
-      burst_rename (&ms, last_uid);
-      mh_msgset_free (&ms);
+      burst_rename (ms, last_uid);
+      mu_msgset_free (ms);
 
       finalize_inplace (last_uid);
 
