@@ -1,6 +1,6 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 1999, 2000, 2001, 2002, 2003, 
-   2005, 2006, 2007 Free Software Foundation, Inc.
+   Copyright (C) 1999-2003, 2005-2007, 2009-2012, 2014-2016 Free
+   Software Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,22 +13,25 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with GNU Mailutils; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
-   MA 02110-1301 USA */
+   along with GNU Mailutils.  If not, see <http://www.gnu.org/licenses/>. */
 
 /* This module implements execution of MH format strings. */
 
 #include <mh.h>
 #include <mh_format.h>
 #include <mailutils/mime.h>
+#include <mailutils/opool.h>
 
-#ifdef HAVE_STRINGS_H
-# include <strings.h>
-#endif
+#include <string.h>
+#include <ctype.h>
+#include "mbiter.h"
+#include "mbchar.h"
+#include "mbswidth.h"
 
 static char *_get_builtin_name (mh_builtin_fp ptr);
 
+#define DFLWIDTH(mach) ((mach)->width - (mach)->ind)
+
 /* Functions for handling string objects. */
 
 void
@@ -51,7 +54,7 @@ strobj_create (strobj_t *lvalue, const char *str)
   else
     {
       lvalue->size = strlen (str) + 1;
-      lvalue->ptr = xmalloc (lvalue->size);
+      lvalue->ptr = mu_alloc (lvalue->size);
       memcpy (lvalue->ptr, str, lvalue->size);
     }
 }
@@ -93,30 +96,66 @@ strobj_realloc (strobj_t *obj, size_t length)
   if (strobj_is_static (obj))
     {
       char *value = strobj_ptr (obj);
-      obj->ptr = xmalloc (length);
+      obj->ptr = mu_alloc (length);
       strncpy (obj->ptr, value, length-1);
       obj->ptr[length-1] = 0;
       obj->size = length;
     }
   else
     {
-      obj->ptr = xrealloc (obj->ptr, length);
+      obj->ptr = mu_realloc (obj->ptr, length);
       obj->ptr[length-1] = 0;
       obj->size = length;
     }
 }
 
-/* Compress whitespace in a string */
+/* Return the length (number of octets) of a substring of
+   string STR of length LEN, such that it contains NCOL
+   multibyte characters. */
+int
+mbsubstrlen (char *str, size_t len, size_t ncol)
+{
+  int ret = 0;
+  mbi_iterator_t iter;
+
+  if (ncol <= 0)
+    return 0;
+  
+  for (mbi_init (iter, str, len);
+       ncol && mbi_avail (iter);
+       ncol--, mbi_advance (iter))
+    ret += mb_len (mbi_cur (iter));
+  return ret;
+}
+
+/* Return the number of multibyte characters in the first LEN bytes
+   of character string STRING.  */
+size_t
+mbsnlen (char *str, size_t len)
+{
+  int ret = 0;
+  mbi_iterator_t iter;
+
+  for (mbi_init (iter, str, len); mbi_avail (iter); mbi_advance (iter))
+    ret++;
+  return ret;
+}
+
+/* Compress whitespace in a string (multi-byte) */
 static void
-compress_ws (char *str, size_t *size)
+compress_ws (char *str, size_t *psize)
 {
   unsigned char *p, *q;
-  size_t len = *size;
+  size_t size = *psize;
+  mbi_iterator_t iter;
   int space = 0;
-
-  for (p = q = (unsigned char*) str; len; len--, q++)
+  
+  for (p = q = (unsigned char*) str,
+	 mbi_init (iter, str, size);
+       mbi_avail (iter);
+       mbi_advance (iter))
     {
-      if (isspace (*q))
+      if (mb_isspace (mbi_cur (iter)))
 	{
 	  if (!space)
 	    *p++ = ' ';
@@ -126,19 +165,32 @@ compress_ws (char *str, size_t *size)
       else if (space)
 	space = 0;
 
-      if (isprint (*q))
-	*p++ = *q;
+      if (mb_isprint (mbi_cur (iter)))
+	{
+	  size_t len = mb_len (mbi_cur (iter));
+	  memcpy (p, mb_ptr (mbi_cur (iter)), len);
+	  p += len;
+	}
     }
   *p = 0;
-  *size = p - (unsigned char*) str;
+  *psize = p - (unsigned char*) str;
 }
 
+#define COMPRESS_WS(mach, str, size)		\
+  do						\
+    {						\
+      if ((mach)->fmtflags & MH_FMT_COMPWS)	\
+	compress_ws (str, size);		\
+    }						\
+  while (0)
+
 static void
 put_string (struct mh_machine *mach, char *str, int len)
 {
   if (len == 0)
-    len = strlen (str);
-  obstack_grow (&mach->stk, str, len);
+    return;
+  mu_opool_append (mach->pool, str, len);
+  len = mbsnwidth (str, len, 0);
   mach->ind += len;
 }
 
@@ -148,40 +200,46 @@ print_hdr_segment (struct mh_machine *mach, char *str, size_t len)
   if (!len)
     len = strlen (str);
 
-  if (len < mach->width)
+  if (mbsnlen (str, len) < mach->width)
     put_string (mach, str, len);
   else
     {
-      char *endp = str + len;
-      
-      while (str < endp)
+      while (1)
 	{
-	  size_t rest;
-	  char *p;
-	  size_t size;
+	  mbi_iterator_t iter;
+	  size_t rest = DFLWIDTH (mach);
+	  size_t width = mbsnlen (str, len);
+	  size_t off, size;
 	  
-	  size = endp - str;
-	  rest = mach->width - mach->ind;
-	  if (size < rest)
+	  if (width <= rest)
 	    {
-	      put_string (mach, str, size);
+	      put_string (mach, str, len);
 	      break;
 	    }
-	  
-	  for (p = str + rest - 1; p > str && !isspace (*p); p--)
-	    ;
 
-	  if (p > str)
+	  size = off = 0;
+	  for (mbi_init (iter, str, len);
+	       mbi_avail (iter);
+	       mbi_advance (iter))
 	    {
-	      put_string (mach, str, p - str);
-	      put_string (mach, "\n\t", 0);
+	      if (mb_isspace (mbi_cur (iter)))
+		off = size;
+	      size += mb_len (mbi_cur (iter));
+	    }
+
+	  if (off > 0)
+	    {
+	      put_string (mach, str, off);
+	      put_string (mach, "\n        ", 9);
 	      mach->ind = 8;
-	      str = p;
+	      str += off;
+	      len -= off;
 	    }
 	  else
 	    {
-	      put_string (mach, str, size);
-	      str += len;
+	      size = mbsubstrlen (str, len, rest);
+	      put_string (mach, str, len);
+	      break;
 	    }
 	}
     }
@@ -224,7 +282,7 @@ print_simple_segment (struct mh_machine *mach, size_t width,
   if (!width)
     width = mach->width;
 
-  rest = width - mach->ind;
+  rest = DFLWIDTH (mach);
   if (rest == 0)
     {
       if (len == 1 && str[0] == '\n')
@@ -232,10 +290,7 @@ print_simple_segment (struct mh_machine *mach, size_t width,
       return;
     }
   
-  if (len > rest)
-    len = rest;
-
-  put_string (mach, str, len);
+  put_string (mach, str, mbsubstrlen (str, len, rest));
 }
 
 static void
@@ -263,26 +318,55 @@ print_string (struct mh_machine *mach, size_t width, char *str)
 }
   
 static void
-print_fmt_string (struct mh_machine *mach, size_t fmtwidth, char *str)
+print_fmt_segment (struct mh_machine *mach, size_t fmtwidth, char *str,
+		   size_t len)
 {
-  size_t len = strlen (str);
-  if (len > fmtwidth)
-    len = fmtwidth;
+  size_t width = mbsnlen (str, len);
+
+  if (fmtwidth && width > fmtwidth)
+    {
+      len = mbsubstrlen (str, len, fmtwidth);
+      width = fmtwidth;
+    }
+  else
+    len = mbsubstrlen (str, len, DFLWIDTH (mach));
+  
   put_string (mach, str, len);
 
-  if (fmtwidth > len)
+  if (fmtwidth > width)
     {
-      fmtwidth -= len;
+      fmtwidth -= width;
       mach->ind += fmtwidth;
       while (fmtwidth--)
-	obstack_1grow (&mach->stk, ' ');
+	mu_opool_append_char (mach->pool, ' ');
     }
 }
 
+static void
+print_fmt_string (struct mh_machine *mach, size_t fmtwidth, char *str)
+{
+  char *p = strchr (str, '\n');
+  while (p)
+    {
+      print_fmt_segment (mach, fmtwidth, str, p - str + 1);
+      mach->ind = 0;
+      str = p + 1;
+      p = strchr (str, '\n');
+    }
+  if (str[0])
+    print_fmt_segment (mach, fmtwidth, str, strlen (str));
+}
+
 static void
 reset_fmt_defaults (struct mh_machine *mach)
 {
+  const char *p;
+  
   mach->fmtflags = 0;
+  p = mh_global_profile_get ("Compress-WS", "yes");
+  if (p && (mu_c_strcasecmp (p, "yes") == 0
+	    || mu_c_strcasecmp (p, "true") == 0))
+    mach->fmtflags |= MH_FMT_COMPWS;
 }
 
 static void
@@ -309,7 +393,7 @@ format_num (struct mh_machine *mach, long num)
 	  ptr = buf;
 	  for (i = n; i < fmtwidth && mach->ind < mach->width;
 	       i++, mach->ind++)
-	    obstack_1grow (&mach->stk, padchar);
+	    mu_opool_append_char (mach->pool, padchar);
 	}
     }
   else
@@ -337,7 +421,7 @@ format_str (struct mh_machine *mach, char *str)
 	  n = fmtwidth - len;
 	  for (i = 0; i < n && mach->ind < mach->width;
 	       i++, mach->ind++, fmtwidth--)
-	    obstack_1grow (&mach->stk, padchar);
+	    mu_opool_append_char (mach->pool, padchar);
 	}
 	      
       print_fmt_string (mach, fmtwidth, str);
@@ -369,7 +453,7 @@ addr_cmp (void *item, void *data)
 static int
 addrlist_lookup (mu_list_t list, mu_address_t addr)
 {
-  return mu_list_do (list, addr_cmp, addr);
+  return mu_list_foreach (list, addr_cmp, addr);
 }
 
 static int
@@ -383,7 +467,7 @@ addr_free (void *item, void *data)
 static void
 addrlist_destroy (mu_list_t *list)
 {
-  mu_list_do (*list, addr_free, NULL);
+  mu_list_foreach (*list, addr_free, NULL);
   mu_list_destroy (list);
 }
 
@@ -395,6 +479,7 @@ mh_format (mh_format_t *fmt, mu_message_t msg, size_t msgno,
 {
   struct mh_machine mach;
   char buf[64];
+  const char *charset = mh_global_profile_get ("Charset", NULL);
   
   memset (&mach, 0, sizeof (mach));
   mach.progsize = fmt->progsize;
@@ -405,15 +490,42 @@ mh_format (mh_format_t *fmt, mu_message_t msg, size_t msgno,
   
   mach.width = width - 1; /* Count the newline */
   mach.pc = 1;
-  obstack_init (&mach.stk);
+  mu_opool_create (&mach.pool, 1);
   mu_list_create (&mach.addrlist);
-
-  reset_fmt_defaults (&mach);
   
-  while (!mach.stop)
+  reset_fmt_defaults (&mach);
+
+#if HAVE_SETLOCALE
+  if (charset && strcmp (charset, "auto"))
+    {
+      /* Try to set LC_CTYPE according to the value of Charset variable.
+	 If Charset is `auto', there's no need to do anything, since it
+	 is already set. Otherwise, we need to construct a valid locale
+	 value with Charset as its codeset part. The problem is, what
+	 language and territory to use for that locale.
+
+	 Neither LANG nor any other environment variable is of any use,
+	 because if it were, the user would have set "Charset: auto".
+	 It would be logical to use 'C' or 'POSIX', but these do not
+	 work with '.UTF-8'. So, in the absence of any viable alternative,
+	 'en_US' is selected. This choice may be overridden by setting
+	 the LC_BASE mh_profile variable to the desired base part.
+      */
+      const char *lc_base = mh_global_profile_get ("LC_BASE", "en_US");
+      char *locale = mu_alloc (strlen (lc_base) + 1 + strlen (charset) + 1);
+      strcpy (locale, lc_base);
+      strcat (locale, ".");
+      strcat (locale, charset);
+      if (!setlocale (LC_CTYPE, locale))
+        mu_error (_("cannot set LC_CTYPE %s"), locale);
+      free (locale);
+    }
+#endif
+  
+  while (!mach.stop && mach.ind < mach.width)
     {
       mh_opcode_t opcode;
-      switch (opcode = MHI_OPCODE(mach.prog[mach.pc++]))
+      switch (opcode = MHI_OPCODE (mach.prog[mach.pc++]))
 	{
 	case mhop_nop:
 	  break;
@@ -423,7 +535,7 @@ mh_format (mh_format_t *fmt, mu_message_t msg, size_t msgno,
 	  break;
 
 	case mhop_branch:
-	  mach.pc += MHI_NUM(mach.prog[mach.pc]);
+	  mach.pc += MHI_NUM (mach.prog[mach.pc]);
 	  break;
 
 	case mhop_num_asgn:
@@ -435,27 +547,27 @@ mh_format (mh_format_t *fmt, mu_message_t msg, size_t msgno,
 	  break;
 	  
 	case mhop_num_arg:
-	  mach.arg_num = MHI_NUM(mach.prog[mach.pc++]);
+	  mach.arg_num = MHI_NUM (mach.prog[mach.pc++]);
 	  break;
 	  
 	case mhop_str_arg:
 	  {
-	    size_t skip = MHI_NUM(mach.prog[mach.pc++]);
-	    strobj_set (&mach.arg_str, MHI_STR(mach.prog[mach.pc]));
+	    size_t skip = MHI_NUM (mach.prog[mach.pc++]);
+	    strobj_set (&mach.arg_str, MHI_STR (mach.prog[mach.pc]));
 	    mach.pc += skip;
 	  }
 	  break;
 
 	case mhop_num_branch:
 	  if (!mach.arg_num)
-	    mach.pc += MHI_NUM(mach.prog[mach.pc]);
+	    mach.pc += MHI_NUM (mach.prog[mach.pc]);
 	  else
 	    mach.pc++;
 	  break;
 
 	case mhop_str_branch:
 	  if (!*strobj_ptr (&mach.arg_str))
-	    mach.pc += MHI_NUM(mach.prog[mach.pc]);
+	    mach.pc += MHI_NUM (mach.prog[mach.pc]);
 	  else
 	    mach.pc++;
 	  break;
@@ -475,7 +587,7 @@ mh_format (mh_format_t *fmt, mu_message_t msg, size_t msgno,
 	      {
 		size_t len = strlen (value);
 		mach.arg_str.size = len + 1;
-      		compress_ws (value, &len);
+      		COMPRESS_WS (&mach, value, &len);
 		mach.arg_str.ptr = value;
 		mach.arg_num = 1;
 	      }
@@ -488,33 +600,34 @@ mh_format (mh_format_t *fmt, mu_message_t msg, size_t msgno,
 	  {
 	    mu_body_t body = NULL;
 	    mu_stream_t stream = NULL;
-	    size_t size = 0, off, str_off, nread;
-	    size_t rest = mach.width - mach.ind;
+	    size_t size = 0, str_off, nread;
+	    size_t rest = DFLWIDTH (&mach);
 
 	    strobj_free (&mach.arg_str);
 	    mu_message_get_body (mach.message, &body);
 	    mu_body_size (body, &size);
-	    mu_body_get_stream (body, &stream);
-	    if (size == 0 || !stream)
+	    if (size == 0)
+	      break;
+	    mu_body_get_streamref (body, &stream);
+	    if (!stream)
 	      break;
 	    if (size > rest)
 	      size = rest;
 
-	    mach.arg_str.ptr = xmalloc (size+1);
+	    mach.arg_str.ptr = mu_alloc (size+1);
 	    mach.arg_str.size = size;
 	    
-	    off = 0;
 	    str_off = 0;
 	    while (!mu_stream_read (stream, mach.arg_str.ptr + str_off,
-				 mach.arg_str.size - str_off, off, &nread)
+				    mach.arg_str.size - str_off, &nread)
 		   && nread != 0
 		   && str_off < size)
 	      {
-		off += nread;
-       		compress_ws (mach.arg_str.ptr + str_off, &nread);
+                COMPRESS_WS (&mach, mach.arg_str.ptr + str_off, &nread);
 		if (nread)
 		  str_off += nread;
 	      }
+	    mu_stream_destroy (&stream);
 	    mach.arg_str.ptr[str_off] = 0;
 	  }
 	  break;
@@ -550,7 +663,7 @@ mh_format (mh_format_t *fmt, mu_message_t msg, size_t msgno,
 	  break;
 
 	case mhop_fmtspec:
-	  mach.fmtflags = MHI_NUM(mach.prog[mach.pc++]);
+	  mach.fmtflags = MHI_NUM (mach.prog[mach.pc++]);
 	  break;
 
 	default:
@@ -564,10 +677,10 @@ mh_format (mh_format_t *fmt, mu_message_t msg, size_t msgno,
   
   if (pret)
     {
-      obstack_1grow (&mach.stk, 0);
-      *pret = strdup (obstack_finish (&mach.stk));
+      mu_opool_append_char (mach.pool, 0);
+      *pret = mu_strdup (mu_opool_finish (mach.pool, NULL));
     }
-  obstack_free (&mach.stk, NULL);
+  mu_opool_destroy (&mach.pool);
   return mach.ind;
 }
 
@@ -601,7 +714,7 @@ mh_format_dump (mh_format_t *fmt)
       int num;
       
       printf ("% 4.4ld: ", (long) pc);
-      switch (opcode = MHI_OPCODE(prog[pc++]))
+      switch (opcode = MHI_OPCODE (prog[pc++]))
 	{
 	case mhop_nop:
 	  printf ("nop");
@@ -613,7 +726,7 @@ mh_format_dump (mh_format_t *fmt)
 	  break;
 
 	case mhop_branch:
-	  num = MHI_NUM(prog[pc++]);
+	  num = MHI_NUM (prog[pc++]);
 	  printf ("branch %d, %lu",
 		  num, (unsigned long) pc + num - 1);
 	  break;
@@ -627,14 +740,14 @@ mh_format_dump (mh_format_t *fmt)
 	  break;
 	  
 	case mhop_num_arg:
-	  num = MHI_NUM(prog[pc++]);
+	  num = MHI_NUM (prog[pc++]);
 	  printf ("num_arg %d", num);
 	  break;
 	  
 	case mhop_str_arg:
 	  {
-	    size_t skip = MHI_NUM(prog[pc++]);
-	    char *s = MHI_STR(prog[pc]);
+	    size_t skip = MHI_NUM (prog[pc++]);
+	    char *s = MHI_STR (prog[pc]);
 	    printf ("str_arg \"");
 	    for (; *s; s++)
 	      {
@@ -682,13 +795,13 @@ mh_format_dump (mh_format_t *fmt)
 	  break;
 
 	case mhop_num_branch:
-	  num = MHI_NUM(prog[pc++]);
+	  num = MHI_NUM (prog[pc++]);
 	  printf ("num_branch %d, %lu",
 		  num, (unsigned long) (pc + num - 1));
 	  break;
 
 	case mhop_str_branch:
-	  num = MHI_NUM(prog[pc++]);
+	  num = MHI_NUM (prog[pc++]);
 	  printf ("str_branch %d, %lu",
 		  num, (unsigned long) (pc + num - 1));
 	  break;
@@ -739,7 +852,7 @@ mh_format_dump (mh_format_t *fmt)
 	  {
 	    int space = 0;
 	    
-	    num = MHI_NUM(prog[pc++]);
+	    num = MHI_NUM (prog[pc++]);
 	    printf ("fmtspec: %#x, ", num);
 	    if (num & MH_FMT_RALIGN)
 	      {
@@ -799,8 +912,19 @@ static void
 builtin_cur (struct mh_machine *mach)
 {
   size_t msgno = mach->msgno;
+  size_t cur;
+  int rc;
+  mu_mailbox_t mbox;
+  
+  rc = mu_message_get_mailbox (mach->message, &mbox);
+  if (rc)
+    {
+      mu_diag_funcall (MU_DIAG_ERROR, "mu_message_get_mailbox", NULL, rc);
+      exit (1);
+    }
   mh_message_number (mach->message, &msgno);
-  mach->arg_num = msgno == current_message;
+  mh_mailbox_get_cur (mbox, &cur); /* FIXME: Cache this */
+  mach->arg_num = msgno == cur;
 }
 
 static void
@@ -826,7 +950,7 @@ builtin_width (struct mh_machine *mach)
 static void
 builtin_charleft (struct mh_machine *mach)
 {
-  mach->arg_num = mach->width - mach->ind;
+  mach->arg_num = DFLWIDTH (mach);
 }
 
 static void
@@ -1022,7 +1146,7 @@ static void
 builtin_putnum (struct mh_machine *mach)
 {
   char *p;
-  asprintf (&p, "%ld", mach->arg_num);
+  mu_asprintf (&p, "%ld", mach->arg_num);
   print_string (mach, 0, p);
   free (p);
 }
@@ -1035,7 +1159,7 @@ builtin_putnumf (struct mh_machine *mach)
 }
 
 static int
-_parse_date (struct mh_machine *mach, struct tm *tm, mu_timezone *tz)
+_parse_date (struct mh_machine *mach, struct tm *tm, struct mu_timezone *tz)
 {
   char *date = strobj_ptr (&mach->arg_str);
   const char *p = date;
@@ -1047,7 +1171,7 @@ _parse_date (struct mh_machine *mach, struct tm *tm, mu_timezone *tz)
       /*mu_error ("can't parse date: [%s]", date);*/
       time (&t);
       *tm = *localtime (&t);
-      tz->utc_offset = mu_utc_offset ();
+      mu_datetime_tz_local (tz);
     }
   
   return 0;
@@ -1058,7 +1182,7 @@ static void
 builtin_sec (struct mh_machine *mach)
 {
   struct tm tm;
-  mu_timezone tz;
+  struct mu_timezone tz;
   
   if (_parse_date (mach, &tm, &tz))
     return;
@@ -1071,7 +1195,7 @@ static void
 builtin_min (struct mh_machine *mach)
 {
   struct tm tm;
-  mu_timezone tz;
+  struct mu_timezone tz;
   
   if (_parse_date (mach, &tm, &tz))
     return;
@@ -1084,7 +1208,7 @@ static void
 builtin_hour (struct mh_machine *mach)
 {
   struct tm tm;
-  mu_timezone tz;
+  struct mu_timezone tz;
   
   if (_parse_date (mach, &tm, &tz))
     return;
@@ -1097,7 +1221,7 @@ static void
 builtin_wday (struct mh_machine *mach)
 {
   struct tm tm;
-  mu_timezone tz;
+  struct mu_timezone tz;
   
   if (_parse_date (mach, &tm, &tz))
     return;
@@ -1110,7 +1234,7 @@ static void
 builtin_day (struct mh_machine *mach)
 {
   struct tm tm;
-  mu_timezone tz;
+  struct mu_timezone tz;
   char buf[80];
   
   if (_parse_date (mach, &tm, &tz))
@@ -1126,7 +1250,7 @@ static void
 builtin_weekday (struct mh_machine *mach)
 {
   struct tm tm;
-  mu_timezone tz;
+  struct mu_timezone tz;
   char buf[80];
   
   if (_parse_date (mach, &tm, &tz))
@@ -1143,7 +1267,7 @@ static void
 builtin_sday (struct mh_machine *mach)
 {
   struct tm tm;
-  mu_timezone tz;
+  struct mu_timezone tz;
 
   /*FIXME: more elaborate check needed */
   if (_parse_date (mach, &tm, &tz))
@@ -1157,7 +1281,7 @@ static void
 builtin_mday (struct mh_machine *mach)
 {
   struct tm tm;
-  mu_timezone tz;
+  struct mu_timezone tz;
   
   if (_parse_date (mach, &tm, &tz))
     return;
@@ -1170,7 +1294,7 @@ static void
 builtin_yday (struct mh_machine *mach)
 {
   struct tm tm;
-  mu_timezone tz;
+  struct mu_timezone tz;
   
   if (_parse_date (mach, &tm, &tz))
     return;
@@ -1183,7 +1307,7 @@ static void
 builtin_mon (struct mh_machine *mach)
 {
   struct tm tm;
-  mu_timezone tz;
+  struct mu_timezone tz;
   
   if (_parse_date (mach, &tm, &tz))
     return;
@@ -1196,7 +1320,7 @@ static void
 builtin_month (struct mh_machine *mach)
 {
   struct tm tm;
-  mu_timezone tz;
+  struct mu_timezone tz;
   char buf[80];
   
   if (_parse_date (mach, &tm, &tz))
@@ -1212,7 +1336,7 @@ static void
 builtin_lmonth (struct mh_machine *mach)
 {
   struct tm tm;
-  mu_timezone tz;
+  struct mu_timezone tz;
   char buf[80];
   
   if (_parse_date (mach, &tm, &tz))
@@ -1228,7 +1352,7 @@ static void
 builtin_year (struct mh_machine *mach)
 {
   struct tm tm;
-  mu_timezone tz;
+  struct mu_timezone tz;
   
   if (_parse_date (mach, &tm, &tz))
     return;
@@ -1241,7 +1365,7 @@ static void
 builtin_zone (struct mh_machine *mach)
 {
   struct tm tm;
-  mu_timezone tz;
+  struct mu_timezone tz;
   
   if (_parse_date (mach, &tm, &tz))
     return;
@@ -1254,7 +1378,7 @@ static void
 builtin_tzone (struct mh_machine *mach)
 {
   struct tm tm;
-  mu_timezone tz;
+  struct mu_timezone tz;
   
   if (_parse_date (mach, &tm, &tz))
     return;
@@ -1285,7 +1409,7 @@ static void
 builtin_szone (struct mh_machine *mach)
 {
   struct tm tm;
-  mu_timezone tz;
+  struct mu_timezone tz;
 
   /*FIXME: more elaborate check needed */
   if (_parse_date (mach, &tm, &tz))
@@ -1313,7 +1437,7 @@ static void
 builtin_dst (struct mh_machine *mach)
 {
   struct tm tm;
-  mu_timezone tz;
+  struct mu_timezone tz;
 
   if (_parse_date (mach, &tm, &tz))
     return;
@@ -1329,11 +1453,11 @@ static void
 builtin_clock (struct mh_machine *mach)
 {
   struct tm tm;
-  mu_timezone tz;
+  struct mu_timezone tz;
 
   if (_parse_date (mach, &tm, &tz))
     return;
-  mach->arg_num = mu_tm2time (&tm, &tz);
+  mach->arg_num = mu_datetime_to_utc (&tm, &tz);
 }
 
 /*     rclock     date     integer  seconds prior to current time*/
@@ -1341,12 +1465,12 @@ void
 builtin_rclock (struct mh_machine *mach)
 {
   struct tm tm;
-  mu_timezone tz;
+  struct mu_timezone tz;
   time_t now = time (NULL);
   
   if (_parse_date (mach, &tm, &tz))
     return;
-  mach->arg_num = now - mu_tm2time (&tm, &tz);
+  mach->arg_num = now - mu_datetime_to_utc (&tm, &tz);
 }
 
 struct
@@ -1368,7 +1492,7 @@ static void
 date_cvt (struct mh_machine *mach, int pretty)
 {
   struct tm tm;
-  mu_timezone tz;
+  struct mu_timezone tz;
   char buf[80];
   int i, len;
   const char *tzname = NULL;
@@ -1442,7 +1566,7 @@ static void
 builtin_nodate (struct mh_machine *mach)
 {
   struct tm tm;
-  mu_timezone tz;
+  struct mu_timezone tz;
   
   mach->arg_num = _parse_date (mach, &tm, &tz);
 }
@@ -1458,7 +1582,20 @@ builtin_proper (struct mh_machine *mach)
 static void
 builtin_friendly (struct mh_machine *mach)
 {
-  /*FIXME: noop*/
+  mu_address_t addr;
+  const char *str;
+  int rc;
+  
+  rc = mu_address_create (&addr, strobj_ptr (&mach->arg_str));
+  if (rc)
+    return;
+
+  if (mu_address_sget_personal (addr, 1, &str) == 0 && str)
+    {
+      strobj_free (&mach->arg_str);
+      strobj_create (&mach->arg_str, str);
+    }
+  mu_address_destroy (&addr);
 }
 
 /*     addr       addr     string   mbox@host or host!mbox rendering*/
@@ -1493,12 +1630,7 @@ builtin_pers (struct mh_machine *mach)
     return;
 
   if (mu_address_sget_personal (addr, 1, &str) == 0 && str)
-    {
-      char *p;
-      asprintf (&p, "\"%s\"", str);
-      strobj_create (&mach->arg_str, p);
-      free (p);
-    }
+    strobj_create (&mach->arg_str, str);
   mu_address_destroy (&addr);
 }
 
@@ -1668,7 +1800,6 @@ static void
 builtin_formataddr (struct mh_machine *mach)
 {
   mu_address_t addr, dest;
-  size_t size;
   int i;
   size_t num;
   const char *buf;
@@ -1704,12 +1835,12 @@ builtin_formataddr (struct mh_machine *mach)
 	}
     }
 
-  if (mu_address_to_string (dest, NULL, 0, &size) == 0)
+  if (mu_address_sget_printable (dest, &buf) == 0)
     {
-      strobj_realloc (&mach->reg_str, size + 1);
-      mu_address_to_string (dest, strobj_ptr (&mach->reg_str), size + 1, NULL);
-      mu_address_destroy (&dest);
+      strobj_realloc (&mach->reg_str, strlen (buf) + 1);
+      strcpy (strobj_ptr (&mach->reg_str), buf);
     }
+  mu_address_destroy (&dest);
 }
 
 /*      putaddr    literal        print str address list with
@@ -1736,8 +1867,8 @@ builtin_unre (struct mh_machine *mach)
   int rc = mu_unre_subject (strobj_ptr (&mach->arg_str), &p);
   if (rc == 0 && p != strobj_ptr (&mach->arg_str))
     {
-      char *q = strdup (p); /* Create a copy, since strobj_create will
-			       destroy p */
+      char *q = mu_strdup (p); /* Create a copy, since strobj_create will
+			          destroy p */
       strobj_free (&mach->arg_str);
       strobj_create (&mach->arg_str, q);
       free (q);
@@ -1814,7 +1945,7 @@ builtin_rcpt (struct mh_machine *mach)
   int rc = mh_decode_rcpt_flag (strobj_ptr (&mach->arg_str));
   if (rc == RCPT_NONE)
     {
-      mu_error (_("Invalid recipient mask"));
+      mu_error (_("invalid recipient mask"));
       /* try to continue anyway */
     }
   mach->arg_num = rc & rcpt_mask;
@@ -1827,7 +1958,7 @@ builtin_concat (struct mh_machine *mach)
 
   if (size == 0)
     return;
-  compress_ws (strobj_ptr (&mach->arg_str), &size);
+  COMPRESS_WS (mach, strobj_ptr (&mach->arg_str), &size);
   if (strobj_len (&mach->reg_str) == 0)
     strobj_copy (&mach->reg_str, &mach->arg_str);
   else
@@ -1851,7 +1982,7 @@ builtin_printhdr (struct mh_machine *mach)
   if (!strobj_is_null (&mach->arg_str))
     {
       s = strobj_len (&mach->arg_str);
-      tmp = strdup (strobj_ptr (&mach->arg_str));
+      tmp = mu_strdup (strobj_ptr (&mach->arg_str));
     }
   
   if (!strobj_is_null (&mach->reg_str))
@@ -1947,10 +2078,10 @@ mh_builtin_t builtin_tab[] = {
   { "comp",     builtin_comp,     mhtype_num,  mhtype_str,  MHA_OPTARG },
   { "compval",  builtin_compval,  mhtype_num,  mhtype_str },	   
   { "trim",     builtin_trim,     mhtype_str,  mhtype_str,  MHA_OPTARG },
-  { "putstr",   builtin_putstr,   mhtype_none,  mhtype_str, MHA_OPTARG },
-  { "putstrf",  builtin_putstrf,  mhtype_none,  mhtype_str, MHA_OPTARG },
-  { "putnum",   builtin_putnum,   mhtype_none,  mhtype_num, MHA_OPTARG },
-  { "putnumf",  builtin_putnumf,  mhtype_none,  mhtype_num, MHA_OPTARG },
+  { "putstr",   builtin_putstr,   mhtype_none, mhtype_str,  MHA_OPTARG },
+  { "putstrf",  builtin_putstrf,  mhtype_none, mhtype_str,  MHA_OPTARG },
+  { "putnum",   builtin_putnum,   mhtype_none, mhtype_num,  MHA_OPTARG },
+  { "putnumf",  builtin_putnumf,  mhtype_none, mhtype_num,  MHA_OPTARG },
   { "sec",      builtin_sec,      mhtype_num,  mhtype_str },
   { "min",      builtin_min,      mhtype_num,  mhtype_str },
   { "hour",     builtin_hour,     mhtype_num,  mhtype_str },
